//=============================================================================
//  MuseScore
//  Linux Music Score Editor
//  $Id:$
//
//  Slash Notation Styles plugin
//
//  Copyright (C)2008-2011 Werner Schweer and others
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License version 2.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//=============================================================================

//
//    This is ECMAScript code (ECMA-262 aka "Java Script")
//

var slashVoice, slashDuration, slashPitch;
var doSlash, doRhythmic, doAccent;
var slashPixmap, rhythmicPixmap, accentPixmap;
var middleLine, AccentLine;
var namedNote; // this is used to translate MIDI pitch into note name

//---------------------------------------------------------
//    init
//    this function will be called on startup of
//    mscore
//---------------------------------------------------------

function init() {
}

// adds a silent, stemless, slash-headed note of specified duration and pitch at cursor position
function addSlash(cursor,duration,pitch) {
  var chord = new Chord();
  chord.tickLen = duration;
  cursor.add(chord);
  var note = new Note();
  chord.addNote(note);
  // some properties seem to work better if set after retrieving chord/note from cursor
  if (cursor.isChord()) {
    chord = cursor.chord();
    chord.noStem = true;
    note = chord.note(0);
    note.pitch = pitch;
    note.noteHead = 5;
    note.velocity = 1; // by using 1 rather than 0, the playback cursor will still follow it
  }
}

// replaces chord with a single silent, stemmed, slash-headed note of specified pitch
function makeGeneric(chord,pitch) {
  for (var i = chord.notes - 1; i > 0; i--) {
    chord.removeNote(i);
  }
  var note = chord.note(0);
  note.pitch = pitch;
  note.noteHead = 5;
  note.velocity = 1; // by using 1 rather than 0, the playback cursor will still follow it
}

// ordinary middle-line rhythmic notation
// cursor points to chord
function makeRhythmic(cursor) {
  var obj = cursor.chord();
  makeGeneric(obj,slashPitch);
}

// small-note, above-staff rhythmic notation
// cursor points to chord or rest
function makeAccent(cursor) {
  var obj;
  if (cursor.isChord()) {
    obj = cursor.chord();
    makeGeneric(obj,slashPitch);
  } else {
    obj = cursor.rest();
  }
  obj.small = true;
}

// loop through music adding slashes or replacing notes with slashes as appropriate
function slashNotationStyles() {

  // get values from dialog
  slashVoice = form.voice.value - 1;
  slashPitch = form.pitch.value;
  slashDuration = form.style.duration.value;

  // determine mode
  doSlash = form.style.doSlash.checked;
  doRhythmic = form.style.doRhythmic.checked;
  doAccent = form.style.doAccent.checked;

  // standard loop through selection
  var cursor       = new Cursor(curScore);
  var selectionEnd = new Cursor(curScore);

  cursor.goToSelectionStart();
  selectionEnd.goToSelectionEnd();
  var startStaff = cursor.staff;
  var endStaff   = selectionEnd.staff;

  // must select something
  if (cursor.eos()) {
    QMessageBox.warning(0,"Invalid Selection","Please select some music and retry operation");
    return;
  }

  curScore.startUndo();

  for (var staff = startStaff; staff < endStaff; ++staff) {

    cursor.goToSelectionStart();
    cursor.voice = slashVoice;
    cursor.staff = staff;

    while (cursor.tick() < selectionEnd.tick()) {

      if (doSlash) {
        addSlash(cursor,slashDuration,slashPitch);
        cursor.next();
        continue;
      }

      if (cursor.isChord()) {
        if (doRhythmic) {
          makeRhythmic(cursor);
        }
        else if (doAccent) {
          makeAccent(cursor);
        }
      } else if (cursor.isRest()) {
        if(doAccent) {
          makeAccent(cursor);
        }
      }
      cursor.next();

    }

  }

  curScore.endUndo();

}

// set the pitchName label in dialog according to specified numeric pitch value
function setPitchName(pitch) {
  var name;
  namedNote.pitch = pitch;
  // for flat keys, respell accidentals with flats
  if (curScore.keysig < 0 && namedNote.tpc >= 20) {
    namedNote.tpc -= 12;
  }
  name = namedNote.name;
  // mark middle C
  if (pitch == 60) {
    name = name+" (middle)"
  }
  form.pitchName.text = name;
}

// set the pitch value in dialog to appropriate value for middle line
function setMiddleLine() {
  form.pitch.value = middleLine;
}

// set the pitch value in dialog to appropriate value for above staff
function setAccentLine() {
  form.pitch.value = accentLine;
}

// set default pitch in dialog to middle line or above staff as appropriate
function setDefaultPitch() {
  if (doSlash || doRhythmic) {
    setMiddleLine();
    form.voice.value = 1;    
  } else if (doAccent) {
    setAccentLine();
    form.voice.value = 3;
  }
}

// set appropriate middle line and above staff values for treble clef
// alter pitch if necessary according to overall score key signature
// set pitch value in dialog to appropriate default
function setTreble() {
  middleLine = 71; // B
  if (curScore.keysig < 0) {
    --middleLine;
  }
  accentLine = 79; // G
  if (curScore.keysig < -4) {
    --accentLine;
  }
  setDefaultPitch();
}

// set appropriate middle line and above staff values for bass clef
// alter pitch if necessary according to overall score key signature
// set pitch value in dialog to appropriate default
function setBass() {
  middleLine = 50; // D
  if (curScore.keysig < -3) {
    --middleLine;
  }
  accentLine = 59; // B
  if (curScore.keysig < 0) {
    --accentLine;
  }
  setDefaultPitch();
}

// display appropriate example for selected style
// set appropriate default pitch & other parameters
function setStyle(checked) {
  if (checked) {
    doSlash = form.style.doSlash.checked;
    doRhythmic = form.style.doRhythmic.checked;
    doAccent = form.style.doAccent.checked;
    if (doSlash) {
      form.example.setPixmap(slashPixmap);
      setMiddleLine();
    } else if (doRhythmic) {
      form.example.setPixmap(rhythmicPixmap);
      setMiddleLine();
    } else if (doAccent) {
      form.example.setPixmap(accentPixmap);
      setAccentLine();
    }
    setDefaultPitch();
  }
}

function run() {

  // read the UI file and create a form out of it
  var loader = new QUiLoader(null);
  var file = new QFile(pluginPath+"/slash.ui");
  file.open(QIODevice.OpenMode(QIODevice.ReadOnly,QIODevice.Text));
  form = loader.load(file,null);

  // connect signals
  form.accepted.connect(slashNotationStyles);
  form.style.doSlash.toggled.connect(setStyle);
  form.style.doRhythmic.toggled.connect(setStyle);
  form.style.doAccent.toggled.connect(setStyle);
  form.pitch['valueChanged(int)'].connect(setPitchName);
  form.clef.treble.clicked.connect(setTreble);
  form.clef.bass.clicked.connect(setBass);

  // initialize global variables and widgets
  namedNote = new Note(curScore);
  slashPixmap = new QPixmap(pluginPath+"/slash.png");
  rhythmicPixmap = new QPixmap(pluginPath+"/rhythmic.png");
  accentPixmap = new QPixmap(pluginPath+"/accent.png");
  doSlash = form.style.doSlash.checked;
  doRhythmic = form.style.doRhythmic.checked;
  doAccent = form.style.doAccent.checked;
  if (form.clef.bass.checked) {
    setBass();
  } else {
    setTreble();
  }
  setStyle(true);

  // show the form
  form.show();

}

//---------------------------------------------------------
//    menu:  defines where the function will be placed
//           in the menu structure
//---------------------------------------------------------

var mscorePlugin = {
      menu: 'Plugins.Lead Sheet.Slash Notation Styles',
      init: init,
      run:  run
}

mscorePlugin;
